mod alignment_test;
mod color_test;
mod colorization;
mod column_names_test;
mod concat_test;
mod disable_test;
mod duplicate_test;
mod extract_test;
mod format_test;
mod formatting_test;
mod height_test;
mod highlingt_test;
mod layout_test;
mod margin_test;
mod merge_test;
mod object_test;
mod padding_test;
mod panel_test;
mod render_settings;
mod reverse_test;
mod rotate_test;
mod shadow_test;
mod span_test;
mod split_test;
mod style_test;
mod theme_test;
mod width_test;
